with Ada.Text_IO; use Ada.Text_IO;
with Ada.Real_Time;

package body Dining_Philosophers is

   --  all forks are initially available
   --+ as denoted by a non-existent holder
   Forks : array (Position_T) of Holder_T := (others => Holder_T'Last);

   --+------ support procedures
   procedure Action (Load : in     Positive);
   function Left (A_Position : Position_T) return Position_T;
   function Right (A_Position : Position_T) return Position_T;

   procedure Action (Load : in     Positive) is
      Factor : constant Positive := 10;
      Dummy  :          Positive := 1;
   begin
      --  this simulates a variable-duration workload
      for I in 1 .. Load loop
         Dummy := Dummy + Factor;
         Dummy := Dummy - Factor;
      end loop;
   end Action;

   function Left (A_Position : Position_T) return Position_T is
   begin
      if (A_Position = Position_T'First) then
         return Position_T'Last;
      else
         return (A_Position - 1);
      end if;
   end Left;

   function Right (A_Position : Position_T) return Position_T is
   begin
      return (A_Position);
   end Right;

   procedure Put_Fork (A_Position : Position_T);

   procedure Put_Fork (A_Position : Position_T) is
   begin
      --  fork becomes available again
      Forks (A_Position) := Holder_T'Last;
   end Put_Fork;

   --+------ end of support procedures

   task body Philosopher is
      use Ada.Real_Time;
      Wait_Time : constant Ada.Real_Time.Time_Span :=
                    Milliseconds (Some_Time);
      --+------ more support procedures
      procedure Think (Think_Load : in     Positive);
      procedure Eat (Eat_Load : in     Positive);
      procedure Take_Fork (A_Position : Position_T;
                           A_Holder   : Holder_T);
      --+-------
      procedure Think (Think_Load : in     Positive) is
         use Ada.Real_Time;
      begin
         --  first seek inspiration
         Action (Think_Load);
         --  then reflect
         delay until Clock + Wait_Time;
      end Think;
      --+-------
      procedure Eat (Eat_Load : in     Positive) is
         use Ada.Real_Time;
      begin
         --  first have a good meal
         Action (Eat_Load);
         --  then take a nap
         delay until Clock + Wait_Time;
      end Eat;
      --+-------
      procedure Take_Fork (A_Position : Position_T;
                           A_Holder   : Holder_T) is
         use Ada.Real_Time;
      begin
         loop
            if (Forks (A_Position) = Holder_T'Last) then
               --  fork is seized and no longer available
               --+ but does this prevent race condition?
               Forks (A_Position) := A_Holder;
               exit;
            else
               --  this alternative should hopefully make chances
               --+ for a deblocking of the situation
               --+ in case the fork was taken
               delay until Clock + Wait_Time;
            end if;
         end loop;
      end Take_Fork;
      --+------ end of support procedures
   begin -- Philosopher
      delay until Clock + Wait_Time;
      loop
         --  first think a good deal
         Think (Think_Load);
         --  then feel hungry and seize left fork
         Take_Fork (Left (My_Position), Holder_T (My_Position));
         --  then seize right fork
         Take_Fork (Right (My_Position), Holder_T (My_Position));
         --  then eat a little
         --+ (a good philosopher eats less than he/she thinks)
         Eat (Think_Load);
         --  then relinquish left fork
         Put_Fork (Left (My_Position));
         --  then relinquish right fork
         Put_Fork (Right (My_Position));
         --  finally inform us that you are done for one round
         Put_Line ("Philosopher " &
                   Natural'Image (Natural (My_Position)) &
                   " has dined. Burp!");
      end loop;
   end Philosopher;

   --  here we declare all the attributes that we need
   --+ to instantiate our philosophers
   Some_Time_1 : constant Positive :=   500;
   Some_Time_2 : constant Positive := 1_000;
   Some_Time_3 : constant Positive := 1_500;
   Some_Time_4 : constant Positive := 2_000;

   Think_Load_1 : constant Positive := 100;
   Think_Load_2 : constant Positive := 200;
   Think_Load_3 : constant Positive := 300;
   Think_Load_4 : constant Positive := 400;

   Position_1 : constant Position_T := Position_T (0);
   Position_2 : constant Position_T := Position_T (1);
   Position_3 : constant Position_T := Position_T (2);
   Position_4 : constant Position_T := Position_T (3);

   --  and, finally, we instantiate our 4 philosophers,
   --+ choosing an arbitrary order of creation
   Philosopher_1 : Philosopher (Position_1, Think_Load_1, Some_Time_1);
   Philosopher_3 : Philosopher (Position_3, Think_Load_3, Some_Time_3);
   Philosopher_2 : Philosopher (Position_2, Think_Load_2, Some_Time_2);
   Philosopher_4 : Philosopher (Position_4, Think_Load_4, Some_Time_4);

   procedure Status_Check (Identity : Positive) is
      --+------ support procedure
      procedure Print_Error (Identity : Positive);
      procedure Print_Error (Identity : Positive) is
      begin
         Put_Line ("Philosopher " &
                   Positive'Image (Identity) & " has terminated!");
      end Print_Error;
      --+------ support procedure
      Error : Boolean := False;
   begin
      case Identity is
         when 1 =>
            if Philosopher_1'Terminated then Error := True;
            end if;
         when 2 =>
            if Philosopher_2'Terminated then Error := True;
            end if;
         when 3 =>
            if Philosopher_3'Terminated then Error := True;
            end if;
         when 4 =>
            if Philosopher_4'Terminated then Error := True;
            end if;
         when others =>
            Put_Line ("Parameter error in procedure Status_Check.");
      end case;
      if Error then Print_Error (Identity);
      end if;
   end Status_Check;

end Dining_Philosophers;
